/*
    DoubleSimple: DoubleKahanSum.m
    Copyright (C) 2003-2006  Macaulay Institute

    This file is part of DoubleSimple, a series of classes for implementing
    various remedies to floating point arithmetic errors.

    DoubleSimple is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DoubleSimple is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*

Implementation of DoubleKahanSum class

*/

#import <DoubleKahanSum.h>
#ifdef INTEL
#  import <CIieeefp.h>
static fp_pctl save_pctl;
#endif

@implementation DoubleKahanSum

/* Methods to override from DoubleSimple */

+newUp {
  DoubleKahanSum *obj = [super newUp];
  obj->correction = 0.0;
  return obj;
}

-set: (DoubleSimple *)value {
  [super set: value];
  correction = 0.0;
  return self;
}

-cset: (double)value {
  [super cset: value];
  correction = 0.0;
  return self;
}

-setStr: (char *)str {
  [super setStr: str];
  correction = 0.0;
  return self;
}

-read: (FILE *)fp {
  [super read: fp];
  correction = 0.0;
  return self;
}

-add: (DoubleSimple *)value {
  double corrected_term, temp_sum;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif
  corrected_term = [value get] - correction;
  temp_sum = d + corrected_term;
  correction = (temp_sum - d) - corrected_term;
  d = temp_sum;
#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  
  return self;
}

-sub: (DoubleSimple *)value {
  double corrected_term, temp_sum;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif
  corrected_term = -[value get] - correction;
  temp_sum = d + corrected_term;
  correction = (temp_sum - d) - corrected_term;
  d = temp_sum;
#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  
  return self;
}


-iadd: (int)value {
  double corrected_term, temp_sum;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif
  corrected_term = (double)value - correction;
  temp_sum = d + corrected_term;
  correction = (temp_sum - d) - corrected_term;
  d = temp_sum;
#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  
  return self;
}

-isub: (int)value {
  double corrected_term, temp_sum;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif
  corrected_term = -(double)value - correction;
  temp_sum = d + corrected_term;
  correction = (temp_sum - d) - corrected_term;
  d = temp_sum;
#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}


-cadd: (double)value {
  double corrected_term, temp_sum;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif
  corrected_term = value - correction;
  temp_sum = d + corrected_term;
  correction = (temp_sum - d) - corrected_term;
  d = temp_sum;
#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
 
  return self;
}

-csub: (double)value {
  double corrected_term, temp_sum;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif
  corrected_term = -value - correction;
  temp_sum = d + corrected_term;
  correction = (temp_sum - d) - corrected_term;
  d = temp_sum;
#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  
  return self;
}

@end
